Require Export FlatCompletePreorder (* required by Fcpo Function *).

Set Implicit Arguments.
Unset Strict Implicit.

Fcpo Function gg (n:nat) : partial nat := Def n.
(* This command immediately generates
   1. the recursive functional f_gg (which is constructed uniquely)
   2. the goal requiring to prove monotonicity of f_gg and continuity of the
      function defined by lifting f_gg to the type of monotonic functions *)
split.
(* left conjunct: monotonicity statement *)
unfold monotonic.
intros x y Hxy.
apply ord_refl_law.
(* right conjunct: continuity statement *)
intros mono_proof c n.
assert (simpl_gg: forall f n,
                    ((@Build_monofun (funord nat (flat_ord nat))
                       (funord nat (flat_ord nat)) f_gg mono_proof) f n)
                  = f_gg f n) by trivial.
rewrite simpl_gg.
destruct (lub_flat_cpo_witness (c <_o> n)) as [m Hwit].
apply ord_trans_law with ((((@Build_monofun (funord nat (flat_ord nat))
                       (funord nat (flat_ord nat)) f_gg mono_proof) @ c) m) n).
rewrite simpl_monofunord_composition (*simpl*).
unfold f_gg.
apply ord_refl_law.
apply (lub_upper_bound_law (((@Build_monofun (funord nat (flat_ord nat))
                       (funord nat (flat_ord nat)) f_gg mono_proof) @ c))).
Qed (* lemma gg_monotonic_continuous *).
(* after "Qed" the command generates the following definitions:

Definition mono_gg := Build_monofun (proj1 gg_monotonic_continuous).

Definition cont_gg := Build_contfun
  ((proj2 gg_monotonic_continuous) (proj1 gg_monotonic_continuous)).

Definition gg := fixp cont_gg.

The function gg is the required representation of the (general recursive)
function in question.
*)

(* the instruction below computes the term which however cannot be evaluated
   to a value in nat *)
Eval compute in (gg 1).

(* load the values for computationally irrelevant axioms *)
Require Export ExtractionConsts.

(* extract gg into OCaml code *)
Extraction "gg" gg.

(* a demonstration of a rewrite lemma *)
Lemma simpl_cont_gg : forall f n, cont_gg f n = mono_gg f n.
reflexivity.
Qed.
